<?php
/**
 * DokuWiki farm management helper
 *
 * @license    GPL 2 (http://www.gnu.org/licenses/gpl.html)
 * @author     Etienne MELEARD <etienne.meleard AT renater.fr>
 */

/**
 * All errors are handled through exception throwing.
 * 
 * Exceptions codes :
 *  Animal creation/delete :
 *      200 : Animal already exists
 *      201 : Template not found
 *      202 : Copy failed
 *  Animal management :
 *      300 : Animal does not exists
 *      301 : ACL update failed
 *      302 : Unknown config type
 *      303 : Local config update failed
 *      304 : Protected config update failed
 *      305 : Page update failed
 *  Animal deletion :
 *      400 : Animal deletion failed
*/
class dw_farm_exception extends Exception {}

/**
 * Animal manager
 */
class dw_farm_animal {
    /**
     * Files extensions placeholder replacment is allowed in (regexp)
     */
    const REPLACE_IN = '/^(txt|conf|php)$/';
    
    /**
     * Checks if an animal exists
     * 
     * @param string $id <animal_id> or <host>/<animal_id> if DOKU_FARM_MULTIHOSTING
     * 
     * @return bool
     */
    public static function exists($id) {
        if(!DOKU_FARM_BARN) return false;
        return @is_dir(DOKU_FARM_BARN.$id);
    }
    
    /**
     * Get existing wikis list
     * 
     * @param string $filter_host used if DOKU_FARM_MULTIHOSTING, empty means "all hosts", wilcard accepted
     * 
     * @return array
     */
    public static function getList($filter_host = null) {
        if($filter_host) $filter_host = '`^'.str_replace('*', '.*', preg_quote($filter_host, '`')).'$`';
        $list = array();
        if(DOKU_FARM_MULTIHOSTING) {
            foreach(scandir(DOKU_FARM_BARN) as $host) {
                if(substr($host, 0, 1) == '.') continue;
                if(!@is_dir(DOKU_FARM_BARN.$host)) continue;
                
                if($filter_host && !preg_match($filter_host, $host)) continue;
                
                foreach(scandir(DOKU_FARM_BARN.$host) as $id) {
                    if(substr($id, 0, 1) == '.') continue;
                    if(substr($id, 0, 1) == '_') continue;
                    $list[$host.'/'.$id] = new self($host.'/'.$id);
                }
            }
            
        } else {
            foreach(scandir(DOKU_FARM_BARN) as $id) {
                if(substr($id, 0, 1) == '.') continue;
                if(substr($id, 0, 1) == '_') continue;
                $list[$id] = new self($id);
            }
        }
        
        return $list;
    }
    
    /**
     * Get statistics about existing wikis
     * 
     * @param string $filter_host used if DOKU_FARM_MULTIHOSTING, empty means "all hosts", wilcard accepted
     * 
     * @return array
     */
    public static function getAllStats($filter_host = null) {
        $stats = array();
        foreach(self::getList($filter_host) as $id => $a) $stats[$id] = $a->getStats();
        return $stats;
    }
    
    /**
     * Recursively copy a wiki structure and data
     * 
     * @param string $from <animal_id> (or <host>/<animal_id> if DOKU_FARM_MULTIHOSTING) of source wiki
     * @param string $to <animal_id> (or <host>/<animal_id> if DOKU_FARM_MULTIHOSTING) of destination wiki
     * @param array $placeholders to replace in data
     * 
     * @return bool
     */
    private static function copyRecursive($from, $to, $placeholders = null) {
        if(@is_file(DOKU_FARM_BARN.$from)) {
            $ext = pathinfo($from, PATHINFO_EXTENSION);
            if(!$placeholders || !preg_match(self::REPLACE_IN, $ext)) return @copy(DOKU_FARM_BARN.$from, DOKU_FARM_BARN.$to);
            
            $content = file_get_contents(DOKU_FARM_BARN.$from);
            foreach($placeholders as $k => $v) $content = preg_replace('`\(%\s*'.$k.'\s*%\)`', $v, $content);
            
            if($fh = fopen(DOKU_FARM_BARN.$to, 'w')) {
                fwrite($fh, $content);
                fclose($fh);
            }else return false;
            
            return true;
        }
        
        if(!@is_dir(DOKU_FARM_BARN.$to) && !mkdir(DOKU_FARM_BARN.$to, 0775, true)) return false;
        
        foreach(scandir(DOKU_FARM_BARN.$from) as $i) if($i != '.' && $i != '..') {
            if(!self::copyRecursive($from.'/'.$i, $to.'/'.$i, $placeholders)) return false;
        }
        
        return true;
    }
    
    /**
     * Recursively remove a wiki
     * 
     * @param string $id <animal_id> or <host>/<animal_id> if DOKU_FARM_MULTIHOSTING
     * 
     * @return bool
     */
    private static function removeRecursive($id) {
        foreach(scandir(DOKU_FARM_BARN.$id) as $i) if($i != '.' && $i != '..') {
            if(@is_file(DOKU_FARM_BARN.$id.'/'.$i) && !@unlink(DOKU_FARM_BARN.$id.'/'.$i)) return false;
            if(@is_dir(DOKU_FARM_BARN.$id.'/'.$i) && !self::removeRecursive($id.'/'.$i.'/')) return false;
        }
        return @rmdir(DOKU_FARM_BARN.$id);
    }
    
    /**
     * Transform a nested config array into a flat one
     * 
     * @param array $conf nested config
     * @param string $prefix array path prefix
     * 
     * @return array
     */
    private static function flattenConf($conf, $prefix = '') {
        $out = array();
        foreach($conf as $k => $v) {
            if(is_array($v)) {
                $v = $this->flattenConf($v, "['".$k."']");
            }else $v = array("['".$k."']" => $v);
            
            foreach($v as $sk => $sv) $out[$prefix.$sk] = $sv;
        }
        return $out;
    }
    
    /**
     * Process config for printing
     * 
     * @param mixed $v value
     * 
     * @return string
     */
    private static function printConf($v) {
        if(is_numeric($v)) return "$v";
        if(is_string($v)) return "'".str_replace("'", '\\\'', $v)."'";
        return "''";
    }
    
    /**
     * Get filesystem statistics
     * 
     * @param string $path to analyse
     * @param bool $collect for intermediate collection
     * 
     * @return array
     */
    private static function getFSStats($path, $collect = false) {
        $fs = array();
        foreach(scandir(DOKU_FARM_BARN.$path) as $i) if($i != '.' && $i != '..') {
            if(@is_file(DOKU_FARM_BARN.$path.'/'.$i)) $fs[$path.'/'.$i] = filesize(DOKU_FARM_BARN.$path.'/'.$i);
            if(@is_dir(DOKU_FARM_BARN.$path.'/'.$i)) $fs = array_merge($fs, self::getFSStats($path.'/'.$i, true));
        }
        
        $c = count($fs);
        $t = array_sum($fs);
        return $collect ? $fs : array(
            'count' => $c,
            'total' => $t,
            'avg' => $c ? $t / $c : 0,
            'min' => $c ? min($fs) : 0,
            'max' => $c ? max($fs) : 0
        );
    }
    
    /**
     * Create a new animal
     * 
     * @param string $id <animal_id> or <host>/<animal_id> if DOKU_FARM_MULTIHOSTING
     * @param string $template <animal_id> or <host>/<animal_id> if DOKU_FARM_MULTIHOSTING
     * @param array $customizations
     * 
     * @return self
     */ 
    public static function create($id, $template, $customizations = array()) {
        if(self::exists($id)) throw new dw_farm_exception('Animal already exists', 200);
        if(!self::exists($template)) throw new dw_farm_exception('Template not found', 201);
        
        $placeholders = array_key_exists('placeholders', $customizations) ? $customizations['placeholders'] : null;
        if(!self::copyRecursive($template, $id, $placeholders)) throw new dw_farm_exception('Copy failed', 202);
        
        $a = new self($id);
        
        if(count($customizations)) $a->edit($customizations);
        
        return $a;
    }
    
    /**
     * Animal id
     */
    private $id = null;
    
    /**
     * Constructor
     * 
     * @param string $id <animal_id> or <host>/<animal_id> if DOKU_FARM_MULTIHOSTING
     * 
     * @return self
     */
    public function __construct($id) {
        if(!self::exists($id)) throw new dw_farm_exception('Animal does not exists', 300);
        
        $this->id = $id;
    }
    
    /**
     * Apply changes
     * 
     * @param array $changes
     * 
     * @return self
     */
    public function edit($changes = array()) {
        foreach($changes as $tg => $dfn) switch($tg) {
            case 'conf' : $this->editConfig($dfn); break;
            case 'acl' : $this->editACL($dfn); break;
            case 'acl_override' : $this->editACL($dfn, true); break;
            case 'page' : foreach($dfn as $ch) $this->editPage($ch['target'], $ch['data'], $ch['searchandreplace']); break;
        }
        return $this;
    }
    
    /**
     * Modify config
     * 
     * @param array $config
     * @param string $target config type
     */
    public function editConfig($config, $target = null) {
        if($target) {
            $handled_errors = array(
                'local' => array(303, 'Local config update failed'),
                'protected' => array(304, 'Protected config update failed')
            );
            if(!array_key_exists($target, $handled_errors)) throw new dw_farm_exception('Unknown config type', 302);
            
            $file = DOKU_FARM_BARN.$this->id.'/conf/'.$target.'.php';
            $out = array();
            $conf = self::flattenConf($config);
            
            foreach(explode("\n", @file_get_contents($file)) as $line) {
                if(preg_match('`^\s*\$conf(\[.+\])(\s*=.+)$`', $line, $m)) {
                    if(array_key_exists($m[1], $conf)) {
                        $out[] = '$conf'.$m[1].' = '.self::printConf($conf[$m[1]]).';';
                        unset($conf[$m[1]]);
                    } else {
                        $out[] = $line;
                    }
                } else {
                    $out[] = $line;
                }
            }
            foreach($conf as $k => $v) $out[] = '$conf'.$k.' = '.self::printConf($v).';';
            
            if($fp = @fopen($file, 'w')) {
                fwrite($fp, implode("\n", $out));
                fclose($fp);
            } else throw new dw_farm_exception($handled_errors[$target][1], $handled_errors[$target][0]);
            
        } else {
            foreach($config as $n => $dfn) $this->editConfig($dfn, $n);
        }
        
        return $this;
    }
    
    /**
     * Modify ACLs
     * 
     * @param array $acl
     * @param bool $override
     * 
     * @return self
     */
    public function editACL($acl, $override = false) {
        $new = array();
        foreach($acl as $p) if(preg_match('`\s*([^\s]+)\s+(@?)([^\s]+)\s+([0-9]+)\s*$`', $p, $m)) {
            $new[$m[1]."\t".$m[2].auth_nameencode($m[3])] = $m[4];
        }
        
        $file = DOKU_FARM_BARN.$this->id.'/conf/acl.auth.php';
        $acl = array();
        if(!$override) {
            foreach(explode("\n", trim(@file_get_contents($file))) as $line) {
                if(preg_match('`\s*([^\s]+)\s+([^\s]+)\s+([0-9]+)\s*$`', $line, $m)) {
                    $acl[] = array($m[1]."\t".$m[2], $m[3]);
                } else $acl[] = $line;
            }
        }
        
        foreach($acl as $k => $v) if(is_array($v) && array_key_exists($v[0], $new)) { // Replace existing entries with new levels
            $acl[$k][1] = $new[$v[0]];
            unset($new[$v[0]]);
        }
        
        foreach($new as $k => $v) $acl[] = array($k, $v); // Add new ones
        
        if($fp = @fopen($file, 'w')) {
            foreach($acl as $l) fwrite($fp, (is_array($l) ? $l[0]."\t".$l[1] : $l)."\n");
            fclose($fp);
        } else throw new dw_farm_exception('ACL update failed', 301);
        
        return $this;
    }
    
    /**
     * Edit page source
     * 
     * @param mixed $target page(s) path
     * @param mixed $data content
     * @param bool $searchandreplace
     * 
     * @return self
     */
    public function editPage($target, $data, $searchandreplace = false) {
        if(!$searchandreplace && !is_string($data)) throw new dw_farm_exception('Page update failed', 305);
        
        if($searchandreplace && !is_array($data)) throw new dw_farm_exception('Page update failed', 305);
        
        if(!is_array($target)) $target = array($target);
        
        foreach($target as $t) {
            $file = wikiFN($t);
            $content = @file_get_contents($file);
            
            if($searchandreplace) {
                $content = str_replace(array_keys($data), array_values($data), $content);
            } else $content = $data;
            
            if($fh = fopen('w', $file)) {
                fwrite($fh, $content);
                fclose($fh);
            } else throw new dw_farm_exception('Page update failed', 305);
        }
    }
    
    /**
     * Get statistics about the animal
     * 
     * @return array
     */
    public function getStats() {
        $stats = array();
        foreach(array('pages', 'media') as $tg) $stats[$tg] = self::getFSStats($this->id.'/data/'.$tg);
        return $stats;
    }
    
    /**
     * Delete the animal
     */
    public function delete() {
        if(!self::removeRecursive($this->id)) throw new dw_farm_exception('Animal deletion failed', 400);
    }
}
