<?php
/**
 * farm management helper
 *
 * @license    GPL 2 (http://www.gnu.org/licenses/gpl.html)
 * @author     Etienne MELEARD <etienne.meleard AT renater.fr>
 */

if(!defined('XMLRPC_CLIENT')) die('No XML-RPC client library');
require_once XMLRPC_CLIENT;

// All errors are handled through exception throwing
class remote_dw_exception extends Exception {}
/* Codes :
 * 	Basic :
 * 		100 : XML-RPC client failed to connect
 * 		101 : No credentials given
 * 		102 : CA certificate not found
 * 		103 : Empty response
 * 	Animal creation/delete :
 * 		200 : Animal already exists
 * 		201 : Template not found
 * 		202 : Copy failed
 * 	Animal management :
 * 		300 : Animal does not exists
 * 		301 : ACL update failed
 * 		302 : Unknown config type
 * 		303 : Local config update failed
 * 		304 : Protected config update failed
 * 		305 : Page update failed
 * 	Animal deletion :
 * 		400 : Animal deletion failed
*/

class remote_dw_farm_animal {
	public static function exists($id) {
		return self::call('exists', $id);
	}
	
	public static function getList() {
		return self::call('getList');
	}
	
	public static function getAllStats() {
		return self::call('getAllStats');
	}
	
	public static function create($id, $template, $customizations = array()) {
		self::call('create', $id, $template, $customizations);
		return new self($id, true);
	}
	
	private $id = null;
	
	public function __construct($id, $nocheck = false) {
		if(!$nocheck && !self::exists($id)) throw new remote_dw_exception('Animal does not exists', 300);
		$this->id = $id;
	}
	
	public function edit($changes = array()) {
		self::call('edit', $this->id, $changes);
		return $this;
	}
	
	public function editConfig($config, $name = null) {
		self::call('editConfig', $this->id, $config, $name);
		return $this;
	}
	
	public function editACL($acl, $override = false) {
		self::call('editACL', $this->id, $acl, $override);
		return $this;
	}
	
	public function getStats() {
		return self::call('getStats', $this->id);
	}
	
	public function delete() {
		return self::call('delete', $this->id);
	}
	
	private static $client = null;
	public static function setupClient($url, $app, $secret, $cacert = null, $debug = null) {
		self::$client = new dw_rpc_client($url, $cacert);
		self::$client->app = $app;
		self::$client->secret = $secret;
		if($debug) self::$client->debug = $debug;
	}
	
	private static function call() {
		$args = func_get_args();
		$args[0] = 'plugin.farm.'.$args[0];
		return call_user_func_array(array(self::$client, 'signedCall'), $args);
	}
}

class dw_rpc_client {
	private $url = null;
	private $cacert = null;
	private $app = null;
	private $secret = null;
	
	private $client = null;
	private $debug = 0;
	
	public function __construct($url, $cacert = null) {
		$this->url = $url;
		if($cacert) $this->__set('cacert', $cacert);
	}
	
	public static function prepare($d) {
		if(is_bool($d)) return new xmlrpcval($d, 'boolean');
		if(is_float($d)) return new xmlrpcval($d, 'float');
		if(is_numeric($d)) return new xmlrpcval($d, 'int');
		if(is_string($d)) return new xmlrpcval($d, 'string');
		if(is_array($d)) {
			$lk = (bool)count(preg_grep('`[^0-9]`', array_keys($d)));
			foreach($d as $k => $sd) $d[$k] = self::prepare($sd);
			return new xmlrpcval($d, (bool)count(preg_grep('`[^0-9]`', array_keys($d))) ? 'struct' : 'array');
		}
		return new xmlrpcval(null, 'null');
	}
	
	private function setupClient($reload = false) {
		if(!$this->client || $reload) {
			if(!$this->url) throw new remote_dw_exception('XML-RPC client failed to connect', 100);
			$this->client = new xmlrpc_client($this->url.'/lib/exe/xmlrpc.php');
			if($this->cacert) $this->client->setCaCertificate($this->cacert);
			if($this->debug) $this->client->setDebug($this->debug);
		}
	}
	
	public function call() {
		$args = func_get_args();
		$method = array_shift($args);
		
		$this->setupClient();
		
		return $this->send($method, $args);
	}
	
	public function signedCall() {
		$args = func_get_args();
		$method = array_shift($args);
		
		if(!$this->app || !$this->secret) throw new remote_dw_exception('XML-RPC client failed to connect', 101);
		
		$this->setupClient();
		
		$data = array();
		foreach($args as $arg) {
			if(is_bool($arg)) $data[] = $arg ? 1 : 0;
			elseif(is_numeric($arg) || is_string($arg)) $data[] = $arg;
			else $data[] = json_encode($arg);
		}
		$signature = hash_hmac('sha1', implode('&', $data), $this->secret);
		array_unshift($args, $signature);
		array_unshift($args, $this->app);
		
		return $this->send($method, $args);
	}
	
	private function send($method, $args) {
		$message = new xmlrpcmsg($method);
		foreach($args as $arg) $message->addParam(self::prepare($arg));
		$response = $this->client->send($message);
		if($response == false) throw new remote_dw_exception('Empty response', 103);
		if($c = $response->faultCode()) {
			if($c >= 99000) { // Custom code
				throw new remote_dw_exception($response->faultString(), $c - 99000);
			}else{ // Backend error
				throw new remote_dw_exception($response->faultString(), $c + 99000);
			}
		}
		
		return php_xmlrpc_decode($response->value());
	}
	
	public function __set($k, $v) {
		switch($k) {
			case 'url' :
				$this->url = $v;
				if($this->client) $this->setupClient(true);
				break;
			case 'cacert' :
				if(!@file_exists($v)) throw new remote_dw_farm_management_exception('CA certificate not found', 102);
				$this->cacert = $v;
				if($this->client) $this->client->setCaCertificate($v);
				break;
			case 'app' :
				$this->app = $v;
				break;
			case 'secret' :
				$this->secret = $v;
				break;
			case 'debug' :
				$this->debug = (int)$v;
				if($this->client) $this->client->setDebug((int)$v);
				break;
		}
	}
}
